import clodius.tiles.format as hgfo
import pandas as pd

import numpy as np
import pandas as pd
import h5py

def csv_to_points(csv_file, output_file):
    '''
    Convert a csv file containing points to a numpy array 
    of [[x,y]] values.

    Parameters:
    -----------
    csv_file: string 
        The filename of the data file 

    '''
    df = pd.read_table(csv_file, delimiter=',')

    min_x = df['x'].min()
    max_x = df['x'].max()
    min_y = df['y'].min()
    max_y = df['y'].max()
    
    width = max_x - min_x
    height = max_y - min_y
    
    max_width = max(width, height)
    # print("max_width:", max_width, min_x, min_y, max_x, max_y)
    max_zoom = 30
    
    with h5py.File(output_file, 'w') as f_out:
        dataset = f_out.create_dataset('values', (len(df), 2), compression='gzip', dtype=np.float32)
        dataset[:] = df.reindex(columns=['x','y']).as_matrix()
        
        dataset.attrs['min_x'] = min_x
        dataset.attrs['max_x'] = max_x
        dataset.attrs['min_y'] = min_y
        dataset.attrs['max_y'] = max_y
        dataset.attrs['max_zoom'] = max_zoom
        dataset.attrs['max_width'] = max_width

    
    info = {
        'min_pos': [min_x, min_y],
        'max_pos': [min_y, max_y],
        'max_zoom': max_zoom,
        'max_width': max_width
    }
    
    return df.reindex(columns=['x', 'y'])

def tileset_info(points_file):
    '''
    Calculate the extent, etc...
    '''
    with h5py.File(points_file, 'r') as f_in:
        dset = f_in['values']
        
        return {
            'min_pos': [float(dset.attrs['min_x']), float(dset.attrs['min_y'])],
            'max_pos': [float(dset.attrs['max_y']), float(dset.attrs['max_y'])],
            'max_width': float(dset.attrs['max_width']),
            'max_zoom': int(dset.attrs['max_zoom']),
            'mirror_tiles': 'false'
        }

def tile_bounds(points_file, z, x, y, width=1, height=1):
    '''
    Get the boundaries of a tile
    
    Parameters:
    -----------
    tileset_info: { min_pos, max_pos, max_width}
        Information about the bounds of this tileset
    '''

    tsinfo = tileset_info(points_file)
    tile_width = tsinfo['max_width'] / 2 ** z
        
    x_start = tsinfo['min_pos'][0] + tile_width * x
    x_end = tsinfo['min_pos'][0] + tile_width * (x+width)
    y_start = tsinfo['min_pos'][1] + tile_width * y
    y_end = tsinfo['min_pos'][1] + tile_width * (y+width)
    
    return (x_start, x_end, y_start, y_end)
    
def filter_points(data, extent):
    '''
    Filter points that are within the extent
    
    Parameters:
    -----------
    data: [[]]
        A 2D numpy array containing x,y values
        
    extent: [x_start, x_end, y_start, y_end]
        The region we want to return points within
        
    Returns
    -------
    data: [[]]
        A 2D numpy array containing x,y values
    '''
    # print("extent:", extent)
    # print("data.shape", data.shape, data[:,0])
    data = data[data[:,0] > extent[0]]
    data = data[data[:,0] < extent[1]]
    
    data = data[data[:,1] > extent[2]]
    data = data[data[:,1] < extent[3]]
        
    return data

def density_tiles(points_file, z, x, y, width=1, height=1):
    '''
    Get a 2D histogram of the given region. If the height and 
    width are specified, then we need to partition this into 
    multiple returned tiles.
    '''
    returns = []
    
    with h5py.File(points_file, 'r') as f:
        # get all the points in the region
        all_points = filter_points(f['values'][:], 
                                        tile_bounds(points_file, z, x, y, 
                                            width, height))

        for i in range(width):
            for j in range(height):
                # filter from the larger subregion
                filtered_points = filter_points(all_points, 
                        tile_bounds(points_file, z, x+i, y+j))
                                                    
                dt = np.histogram2d(filtered_points[:,0], 
                                      filtered_points[:,1], bins=256)[0].T
                dt[dt  == 0.] = np.nan
                
                returns += [((z, x+i, y+j), dt)]
                
        return returns

def tiles(points_file, z, x, y, width=1, height=1):
    return [(tile_position, hgfo.format_dense_tile(data.flatten())) for 
            (tile_position, data) in density_tiles(points_file, z, x, y, width, height)]



import collections as col
import sqlite3 

def get_2d_tileset_info(db_file):
    conn = sqlite3.connect(db_file)
    c = conn.cursor()

    row = c.execute("SELECT * from tileset_info").fetchone()
    tileset_info = {
        'zoom_step': row[0],
        'max_length': row[1],
        'assembly': row[2],
        'chrom_names': row[3],
        'chrom_sizes': row[4],
        'tile_size': row[5],
        'max_zoom': row[6],
        'max_width': row[7],
        'min_pos': [1, 1],
        'max_pos': [row[1], row[1]]
    }
    conn.close()

    return tileset_info

def get_1D_tiles(db_file, tile_ids):
    '''
    Retrieve a contiguous set of tiles from a 2D db tile file.

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    zoom: int
        The zoom level
    tile_x_pos: int
        The x position of the first tile
    numx: int
        The width of the block of tiles to retrieve

    Returns
    -------
    tiles: {pos: tile_value}
        A set of tiles, indexed by position
    '''
    tileset_info = get_2d_tileset_info(db_file)

    conn = sqlite3.connect(db_file)

    c = conn.cursor()
    tile_width = tileset_info['max_width'] / 2 ** zoom

    tile_x_start_pos = tile_width * tile_x_pos
    tile_x_end_pos = tile_x_start_pos + (numx * tile_width)

    # print('tile_x_start:', tile_x_start_pos, tile_x_end_pos)

    query = '''
    SELECT fromX, toX, fromY, toY, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rToX >= {} AND
        rFromX <= {}
    '''.format(
        zoom,
        tile_x_start_pos,
        tile_x_end_pos,
    )

    rows = c.execute(query).fetchall()

    new_rows = col.defaultdict(list)
    # print("len(rows)", len(rows))

    for r in rows:
        try:
            uid = r[7].decode('utf-8')
        except AttributeError:
            uid = r[7]

        x_start = r[0]
        x_end = r[1]
        y_start = r[2]
        y_end = r[3]

        for i in range(tile_x_pos, tile_x_pos + numx):
            tile_x_start = i * tile_width
            tile_x_end = (i+1) * tile_width

            if (
                x_start < tile_x_end and
                x_end >= tile_x_start
            ):
                # add the position offset to the returned values
                new_rows[i] += [
                    {'xStart': r[0],
                     'xEnd': r[1],
                     'yStart': r[2],
                     'yEnd': r[3],
                     'chrOffset': r[4],
                     'importance': r[5],
                     'uid': uid,
                     'fields': r[6].split('\t')}]
    conn.close()

    return new_rows

def get_2D_tiles(db_file, zoom, tile_x_pos, tile_y_pos, numx=1, numy=1):
    '''
    Retrieve a contiguous set of tiles from a 2D db tile file.

    Parameters
    ----------
    db_file: str
        The filename of the sqlite db file
    zoom: int
        The zoom level
    tile_x_pos: int
        The x position of the first tile
    tile_y_pos: int
        The y position of the first tile
    numx: int
        The width of the block of tiles to retrieve
    numy: int
        The height of the block of tiles to retrieve

    Returns
    -------
    tiles: {pos: tile_value}
        A set of tiles, indexed by position
    '''
    tileset_info = get_2d_tileset_info(db_file)

    conn = sqlite3.connect(db_file)

    c = conn.cursor()
    tile_width = tileset_info['max_width'] / 2 ** zoom

    tile_x_start_pos = tile_width * tile_x_pos
    tile_x_end_pos = tile_x_start_pos + (numx * tile_width)

    tile_y_start_pos = tile_width * tile_y_pos
    tile_y_end_pos = tile_y_start_pos + (numy * tile_width)

    query = '''
    SELECT fromX, toX, fromY, toY, chrOffset, importance, fields, uid
    FROM intervals,position_index
    WHERE
        intervals.id=position_index.id AND
        zoomLevel <= {} AND
        rToX >= {} AND
        rFromX <= {} AND
        rToY >= {} AND
        rFromY <= {}
    '''.format(
        zoom,
        tile_x_start_pos,
        tile_x_end_pos,
        tile_y_start_pos,
        tile_y_end_pos
    )

    rows = c.execute(query).fetchall()

    new_rows = col.defaultdict(list)

    for r in rows:
        try:
            uid = r[7].decode('utf-8')
        except AttributeError:
            uid = r[7]

        x_start = r[0]
        x_end = r[1]
        y_start = r[2]
        y_end = r[3]

        for i in range(tile_x_pos, tile_x_pos + numx):
            for j in range(tile_y_pos, tile_y_pos + numy):
                tile_x_start = i * tile_width
                tile_x_end = (i+1) * tile_width

                tile_y_start = j * tile_width
                tile_y_end = (j+1) * tile_width

                if (
                    x_start < tile_x_end and
                    x_end >= tile_x_start and
                    y_start < tile_y_end and
                    y_end >= tile_y_start
                ):
                    # add the position offset to the returned values
                    new_rows[(i, j)] += [
                        {'xStart': r[0],
                         'xEnd': r[1],
                         'yStart': r[2],
                         'yEnd': r[3],
                         'chrOffset': r[4],
                         'importance': r[5],
                         'uid': uid,
                         'fields': r[6].split('\t')}]
    conn.close()

    return new_rows



import math
import numpy as np
import clodius.tiles.format as ctf
import clodius.tiles.npvector as ctn

import json

def tiles(array, z, x, importances, tile_size=16):
    '''
    Return tiles from the array. If importances are provided,
    then they will be used to prioritize entries. Otherwise we'll
    assign random importances to the entire array.
    '''
    max_zoom, x_start, x_end = ctn.max_zoom_and_data_bounds(array, z, x, tile_size)
    # print("XXXX x_start:x_end", x_start, x_end)
    
    tile_array = array[x_start:x_end]
    tile_importances = importances[x_start:x_end]

    indeces = np.argsort(tile_importances)[::-1][:tile_size]
    
    return [{'x': x, 'label': label, 'importance': importance} for x, label, importance 
        in zip([int(i) for i in x_start + indeces], tile_array[indeces], tile_importances[indeces] )]

def tiles_wrapper(array, tile_ids, importances):
    tile_values = []
    
    for tile_id in tile_ids:
        parts = tile_id.split('.')

        if len(parts) < 3:
            raise IndexError("Not enough tile info present")

        uid = parts[0]
        z = int(parts[1])
        x = int(parts[2])
    
        ret_array = tiles(array, z, x, importances)
        
        tile_values +=  [(tile_id, ret_array)]

    return tile_values


import os.path as op

def infer_filetype(filename):
    _,ext = op.splitext(filename)

    if ext.lower() == '.bw' or ext.lower() == '.bigwig':
        return 'bigwig'
    elif ext.lower() == '.mcool' or ext.lower() == '.cool':
        return 'cooler'
    elif ext.lower() == '.htime':
        return 'time-interval-json'
    elif ext.lower() == '.hitile':
        return 'hitile'

    return None

def infer_datatype(filetype):
    if filetype == 'cooler':
        return 'matrix'
    if filetype == 'bigwig':
        return 'vector'
    if filetype == 'time-interval-json':
        return 'time-interval'
    if filetype == 'hitile':
        return 'vector'



import clodius.tiles.utils as hgut
import json

def tileset_info(df, x_column, y_column):
    minx = df[x_column].min()
    maxx = df[x_column].max()
    
    miny = df[y_column].min()
    maxy = df[y_column].max()
    
    return {
        'min_pos': [minx, miny],
        'max_pos': [maxx, maxy],
        'max_width': max(maxy - miny, maxx - minx),
        'max_zoom': 100
    }
    
def format_data(tiles):
    new_out = []
    
    for tile_id, tile_data in tiles:
        #print("len:", len(tile_data))
        #tile_data['data'] = tile_data['gene1']
        new_out += [(tile_id, json.loads(tile_data.to_json(orient='records')))]
        
    return new_out    

def tiles(df, x_column, y_column, tileset_info, z, x, y, width=1, height=1):
    [minx, miny, maxx, maxy] = hgut.tile_bounds(tileset_info, z, x, y, width, height)
    #print("x", x, 'y', y, width, height)
    max_per_tile = 30
    
    # get the entire region
    df = df.query('{} < {} & {} < {} & {} < {} & {} < {}'.
                    format(minx, x_column,
                          x_column, maxx,
                          miny, y_column,
                          y_column, maxy))
    
    tiles = []
    if width > 1 or height > 1:
        for i in range(width):
            for j in range(height):
                #print("i:", i, "j:", j)
                [minx, miny, maxx, maxy] = hgut.tile_bounds(tileset_info, z, x + i, y + j)

                data = df.query('{} < {} & {} < {} & {} < {} & {} < {}'.
                        format(minx, x_column,
                              x_column, maxx,
                              miny, y_column,
                              y_column, maxy))
                tiles += [((z,x+i,y+j), data[:max_per_tile])]
    else:
        return [((z,x,y), df[:max_per_tile])]
    
    return tiles
                  
